unit CommandLineParserU;

interface

uses
  Classes;

function Parameters: TStringList; overload;
function Parameters(const aCommandLine: string; const KeepOriginal: Boolean = true): TStringList; overload;

function ParameterByIndex(const Index: Integer): string;
function ParameterByName(const Name: string): string;
function ParameterByNameDef(const Name: string; Default: string): string; overload;
function ParameterByNameDef(const Name: string; Default: Integer): Integer; overload;
function ParameterByNameDef(const Name: string; Default: Boolean): Boolean; overload;

function ParameterArgChars(AArgChars: string): string;

implementation

uses
  Windows, Sysutils, StrUtils;

{$IF CompilerVersion < 20.0}
// 20.0 = Delphi 2009
type
  AnsiChar = Char;
  TSysCharSet = set of AnsiChar;

function CharInSet(C: AnsiChar; const CharSet: TSysCharSet): Boolean;
begin
  Result := C in CharSet;
end;
{$IFEND}

type
  ECommandParser = class(Exception);

  TCommandLineParser = class
  private
    FCaseSensitive: Boolean;
    FArgChars: string;
    FDefSwitch: Char;
    FSwitchChars: string;
    FArguments: TStrings;
    FParameters: TStringList;
    FCommandLineToParse: string;
    FKeepOriginal: Boolean;
    procedure SetSwitchChars(const Value: string);
    procedure SetCaseSensitive(const Value: Boolean);
    function GetParameters: TStringList;
    procedure SetCommandLineToParse(const Value: string);
    procedure SetKeepOriginal(const Value: Boolean);
  protected
    procedure InternalProcessCommandLine(ACommandText: string; ACommandList: TStrings);

    procedure ProcessCommandLine;
    procedure ProcessOption(const AOption: string);
  public
    property Parameters: TStringList read GetParameters;
    property CaseSensitive: Boolean read FCaseSensitive write SetCaseSensitive default False;
    property SwitchChars: string read FSwitchChars write SetSwitchChars;
    property ArgChars: string read FArgChars write FArgChars;
    property CommandLineToParse: string read FCommandLineToParse write SetCommandLineToParse;
    property KeepOriginal: Boolean read FKeepOriginal write SetKeepOriginal;

    constructor Create; overload;
    destructor Destroy; override;
  end;

const
  SValSep = '|';
  SSwitchCharDefaults = '/-';
  SArgCharDefaults = '=:';
  QuoteSet = ['"', #39];

var
  CommandParser: TCommandLineParser;

procedure TCommandLineParser.InternalProcessCommandLine(ACommandText: string; ACommandList: TStrings);
var
  I: Integer;
  L: Integer;
  LastPos: Integer;
  Temp: string;
  CurChar, Quote: Char;

  function InQuote: Boolean;
  begin
    Result := Quote <> #0;
  end;

begin
  if ACommandList = nil then
    raise ECommandParser.Create('CommandList is NIL!');

  LastPos := 1;
  Quote := #0;

  I := 1;
  L := Length(ACommandText);
  while I <= L do
  begin
    CurChar := ACommandText[I];
    if CharInSet(CurChar, QuoteSet) then
    begin
      if (CurChar = Quote) then // Implicitly InQuote
        Quote := #0
      else
        Quote := CurChar;
    end
    else if (CharInSet(CurChar, [#10, #13, ' '])) and (not InQuote) then
    begin
      Temp := Trim(Copy(ACommandText, LastPos, I - LastPos));
      if Length(Temp) > 0 then
        ACommandList.Add(Temp);
      LastPos := I + 1;
    end;

    Inc(I);
  end;

  if LastPos < I then
  begin
    Temp := Trim(Copy(ACommandText, LastPos, I - LastPos));
    ACommandList.Add(Temp);
  end;
end;

{ TCommandParser }

constructor TCommandLineParser.Create;
begin
  inherited;
  FParameters := TStringList.Create;
  FParameters.CaseSensitive := False;
  CaseSensitive := False;
  SetSwitchChars(SSwitchCharDefaults);
  FArgChars := SArgCharDefaults;
  FArguments := TStringList.Create;
  ProcessCommandLine;
end;

destructor TCommandLineParser.Destroy;
begin
  FreeAndNil(FParameters);
  FreeAndNil(FArguments);
  inherited;
end;

function TCommandLineParser.GetParameters: TStringList;
begin
  if FParameters = nil then
  begin
    FParameters := TStringList.Create;
    FParameters.CaseSensitive := False;
    ProcessCommandLine;
  end;

  Result := FParameters;
end;

procedure TCommandLineParser.ProcessCommandLine;
var
  I: Integer;
  Option: string;
begin
  FArguments.Clear;

  if FCommandLineToParse = '' then
    FCommandLineToParse := GetCommandLine;

  InternalProcessCommandLine(FCommandLineToParse, FArguments);

  // Get rid of executable name
  if SameText(FArguments[0], '"' + ParamStr(0) + '"') then
    FArguments.Delete(0);

  I := 0;
  while I < FArguments.Count do
  begin
    Option := FArguments[I];
    ProcessOption(Option);
    Inc(I);
  end;
end;

procedure TCommandLineParser.ProcessOption(const AOption: string);
var
  I, J, ArgPos: Integer;
  Opt, ArgValue: string;

  function StringIndex(const ASubText, AText: string): Integer;
  begin
    if CaseSensitive then
      Result := Pos(ASubText, AText)
    else
      Result := Pos(AnsiUppercase(ASubText), AnsiUppercase(AText));
  end;

begin
  Opt := AOption;

  // Opt := Copy(Opt, 2);
  ArgValue := '';
  ArgPos := 0;

  // Look for argument specifier
  J := Length(FArgChars);
  for I := 1 to J do
  begin
    ArgPos := StringIndex(FArgChars[I], Opt);
    if ArgPos > 1 then
      break; // found ArgPos
  end;

  if ArgPos < 1 then // not found a value
    Exit;

  if KeepOriginal then
  begin
    I := FParameters.IndexOfName(Opt);

    if I < 0 then
      FParameters.Add(Opt)
    else
      FParameters[I] := Opt;

    Exit;
  end;

  ArgValue := Copy(Opt, ArgPos + 1);
  Opt := Copy(Opt, 1, ArgPos - 1);

  if (Opt <> '') and (Pos(Opt[1], SSwitchCharDefaults) = 1) then
    Delete(Opt, 1, 1);
  // SSwitchCharDefaults

  // J := Length(FArgChars);
  I := FParameters.IndexOfName(Opt);

  if Length(ArgValue) >= 2 then
  begin
    if CharInSet(ArgValue[1], QuoteSet) then
      if CharInSet(ArgValue[Length(ArgValue)], QuoteSet) then
      begin
        Delete(ArgValue, 1, 1);
        Delete(ArgValue, Length(ArgValue), 1);
      end;
  end;

  if I < 0 then
    FParameters.Add(Opt + FParameters.NameValueSeparator + ArgValue)
  else
    FParameters[I] := Opt + FParameters.NameValueSeparator + ArgValue;
end;

procedure TCommandLineParser.SetCaseSensitive(const Value: Boolean);
begin
  FCaseSensitive := Value;
  FParameters.CaseSensitive := Value;
end;

procedure TCommandLineParser.SetCommandLineToParse(const Value: string);
begin
  FCommandLineToParse := Value;
  FreeAndNil(FParameters);
end;

procedure TCommandLineParser.SetKeepOriginal(const Value: Boolean);
begin
  FKeepOriginal := Value;
  FreeAndNil(FParameters);
end;

procedure TCommandLineParser.SetSwitchChars(const Value: string);
begin
  FSwitchChars := Value;
  if Length(FSwitchChars) > 0 then
    FDefSwitch := FSwitchChars[1]
  else
    FDefSwitch := '/';
end;

function Parameters: TStringList;
begin
  Result := TStringList.Create;
  Result.Assign(CommandParser.Parameters);
end;

function Parameters(const aCommandLine: string; const KeepOriginal: Boolean): TStringList;
begin
  Result := TStringList.Create;
  CommandParser.CommandLineToParse := aCommandLine;
  CommandParser.KeepOriginal := KeepOriginal;
  Result.Assign(CommandParser.Parameters);
end;

function ParameterByIndex(const Index: Integer): string;
begin
  Result := '';

  if not index in [0 .. CommandParser.Parameters.Count - 1] then
    Exit;

  Result := CommandParser.Parameters[Index];
end;

function ParameterByName(const Name: string): string;
begin
  Result := CommandParser.Parameters.Values[Name];
end;

function ParameterByNameDef(const Name: string; Default: string): string;
begin
  Result := ParameterByName(Name);
  if Result = '' then
    Result := Default;
end;

function ParameterByNameDef(const Name: string; Default: Integer): Integer; overload;
begin
  Result := StrToIntDef(ParameterByName(Name), Default);
end;

function ParameterByNameDef(const Name: string; Default: Boolean): Boolean; overload;
begin
  Result := StrToBoolDef(ParameterByName(Name), Default);
end;

function ParameterArgChars(AArgChars: string): string;
begin
  CommandParser.ArgChars := AArgChars;
  CommandParser.ProcessCommandLine;
end;

initialization

CommandParser := TCommandLineParser.Create;

finalization

CommandParser.Free;

end.
